@@ -5,7 +5,7 @@ from .core import IPPToolWrapper
 config = get_config()
 
 wrapper = IPPToolWrapper(config)
-create_job_subscriptions = wrapper.create_job_subscriptions
+create_job_subscription = wrapper.create_job_subscription
 create_printer_subscription = wrapper. create_printer_subscription
 cancel_job = wrapper.cancel_job
 release_job = wrapper.release_job
@@ -24,6 +24,7 @@ get_job_attributes = wrapper.get_job_attributes
 get_jobs = wrapper.get_jobs
 get_printer_attributes = wrapper.get_printer_attributes
 get_subscriptions = wrapper.get_subscriptions
+get_notifications = wrapper.get_notifications
 pause_printer = wrapper.pause_printer
 resume_printer = wrapper.resume_printer
 hold_new_jobs = wrapper.hold_new_jobs
