@@ -36,27 +36,71 @@ from .forms import (cancel_job_form,
                     )
 
 try:
-    if os.getenv('NO_TORNADO', '').lower() in ('1', 'yes', 'true', 't'):
-        raise ImportError
     from tornado.gen import coroutine, Return, Task
     from tornado.ioloop import TimeoutError
 except ImportError:
-    HAS_TORNADO = False
-
     def coroutine(f):
         return f
 
     class TimeoutError(Exception):
         pass
-else:
-    HAS_TORNADO = True
+
+    class Return(Exception):
+        def __init__(self, value):
+            self.value = value
+
+
+def pyipptool_coroutine(method):
+    """
+    Mark the method as a coroutine.
+    If use with tornado the side effect of this decorator will be
+    cancelled and the original_method will be wrapped by
+    tornado.gen.coroutine .
+    Otherwise the sync_coroutine_consumer wrapper
+    will take care to consume the generator synchronously.
+    """
+    method.ipptool_caller = True
+    @functools.wraps(method)
+    def sync_coroutine_consumer(*args, **kw):
+        gen = method(*args, **kw)
+        while True:
+            value = gen.next()
+            try:
+                gen.send(value)
+            except Return as returned:
+                return returned.value
+            except StopIteration:
+                return value
+    sync_coroutine_consumer.original_method = method
+    return sync_coroutine_consumer
+
+
+class MetaAsyncShifter(type):
+    """
+    Based on async flage defined on IPPToolWrapper
+    methods will be decorated by tornado.gen.coroutine otherwise
+    with a fake one.
+    """
+    def __new__(cls, name, bases, attrs):
+        klass = super(MetaAsyncShifter, cls).__new__(cls, name, bases, attrs)
+        if attrs.get('async'):
+            # ASYNC Wrapper
+            for method_name in dir(bases[0]):
+                method = getattr(bases[0], method_name)
+                if getattr(method, 'ipptool_caller', False):
+                    # Patch Method with tornado.gen.coroutine
+                    setattr(klass, method_name,
+                            coroutine(method.original_method))
+        return klass
+
 
 
 class IPPToolWrapper(object):
+    __metaclass__ = MetaAsyncShifter
+    async = False
 
-    def __init__(self, config, io_loop=None):
+    def __init__(self, config):
         self.config = config
-        self.io_loop = io_loop
 
     def authenticate_uri(self, uri):
         if 'login' in self.config and 'password' in self.config:
@@ -83,43 +127,8 @@ class IPPToolWrapper(object):
                 break
             time.sleep(.1)
 
+
     def _call_ipptool(self, uri, request):
-        if HAS_TORNADO:
-            return self._async_call_ipptool(uri, request)
-        return self._sync_call_ipptool(uri, request)
-
-    @coroutine
-    def cleanup_fd(self, name):
-        try:
-            os.unlink(name)
-        except OSError:
-            pass
-
-    @coroutine
-    def _async_call_ipptool(self, uri, request):
-        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
-            temp_file.write(request)
-        from tornado.process import Subprocess
-        process = Subprocess([self.config['ipptool_path'],
-                              self.authenticate_uri(uri), '-X',
-                              temp_file.name],
-                             stdin=subprocess.PIPE,
-                             stdout=Subprocess.STREAM,
-                             stderr=Subprocess.STREAM,
-                             io_loop=self.io_loop)
-        future = []
-        self.io_loop.add_timeout(self.io_loop.time() + self.config['timeout'],
-                                 functools.partial(self.timeout_handler,
-                                                   process.proc, future))
-        stdout, stderr = yield [Task(process.stdout.read_until_close),
-                                Task(process.stderr.read_until_close)]
-        if future:
-            raise TimeoutError
-        yield self.cleanup_fd(temp_file.name)
-
-        raise Return(plistlib.readPlistFromString(stdout))
-
-    def _sync_call_ipptool(self, uri, request):
         with tempfile.NamedTemporaryFile(delete=False) as temp_file:
             temp_file.write(request)
         process = subprocess.Popen([self.config['ipptool_path'],
@@ -140,8 +149,9 @@ class IPPToolWrapper(object):
         timer.cancel()
         if future:
             raise TimeoutError
-        return plistlib.readPlistFromString(stdout)
+        return plistlib.readPlistFromString(stdout)['Tests'][0]
 
+    @pyipptool_coroutine
     def release_job(self, uri,
                     printer_uri=colander.null,
                     job_id=colander.null,
@@ -151,9 +161,10 @@ class IPPToolWrapper(object):
                           'job_id': job_id,
                           'job_uri': job_uri}}}
         request = release_job_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def cancel_job(self, uri,
                    printer_uri=colander.null,
                    job_id=colander.null,
@@ -165,9 +176,10 @@ class IPPToolWrapper(object):
                           'job_uri': job_uri,
                           'purge_job': purge_job}}}
         request = cancel_job_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def create_job_subscription(self,
                                 uri,
                                 printer_uri=None,
@@ -195,9 +207,10 @@ class IPPToolWrapper(object):
               'notify_natural_language': notify_natural_language,
               'notify_time_interval': notify_time_interval}
         request = create_job_subscription_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def create_printer_subscription(
             self,
             uri,
@@ -226,9 +239,10 @@ class IPPToolWrapper(object):
               'notify_lease_duration': notify_lease_duration,
               'notify_time_interval': notify_time_interval}
         request = create_printer_subscription_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def cups_add_modify_printer(self, uri,
                                 printer_uri=None,
                                 auth_info_required=colander.null,
@@ -264,9 +278,10 @@ class IPPToolWrapper(object):
               }
 
         request = cups_add_modify_printer_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def cups_add_modify_class(self, uri,
                               printer_uri=None,
                               auth_info_required=colander.null,
@@ -296,21 +311,24 @@ class IPPToolWrapper(object):
               }
 
         request = cups_add_modify_class_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def cups_delete_printer(self, uri, printer_uri=None):
         kw = {'header': {'operation_attributes': {'printer_uri': printer_uri}}}
         request = cups_delete_printer_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def cups_delete_class(self, uri, printer_uri=None):
         kw = {'header': {'operation_attributes': {'printer_uri': printer_uri}}}
         request = cups_delete_class_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def cups_get_classes(self, uri,
                          first_printer_name=colander.null,
                          limit=colander.null,
@@ -328,9 +346,10 @@ class IPPToolWrapper(object):
                          'requested_attributes': requested_attributes,
                          'requested_user_name': requested_user_name}}}
         request = cups_get_classes_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def cups_get_devices(self, uri,
                          device_class=colander.null,
                          exclude_schemes=colander.null,
@@ -346,9 +365,10 @@ class IPPToolWrapper(object):
                          'requested_attributes': requested_attributes,
                          'timeout': timeout}}}
         request = cups_get_devices_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def cups_get_ppds(self, uri,
                       exclude_schemes=colander.null,
                       include_schemes=colander.null,
@@ -375,9 +395,10 @@ class IPPToolWrapper(object):
                          'requested_attributes': requested_attributes
                          }}}
         request = cups_get_ppds_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def cups_get_printers(self, uri,
                           first_printer_name=colander.null,
                           limit=colander.null,
@@ -395,9 +416,10 @@ class IPPToolWrapper(object):
                          'requested_attributes': requested_attributes,
                          'requested_user_name': requested_user_name}}}
         request = cups_get_printers_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def cups_move_job(self, uri,
                       printer_uri=colander.null,
                       job_id=colander.null,
@@ -411,9 +433,10 @@ class IPPToolWrapper(object):
               'job_printer_uri': job_printer_uri,
               'printer_state_message': printer_state_message}
         request = cups_move_job_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def cups_reject_jobs(self, uri,
                          printer_uri=None,
                          requesting_user_name=None,
@@ -423,9 +446,10 @@ class IPPToolWrapper(object):
                          'requesting_user_name': requesting_user_name}},
               'printer_state_message': printer_state_message}
         request = cups_reject_jobs_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def get_job_attributes(self, uri,
                            printer_uri=colander.null,
                            job_id=colander.null,
@@ -439,9 +463,10 @@ class IPPToolWrapper(object):
                          'requesting_user_name': requesting_user_name,
                          'requested_attributes': requested_attributes}}}
         request = get_job_attributes_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def get_jobs(self, uri,
                  printer_uri=None,
                  requesting_user_name=colander.null,
@@ -457,9 +482,10 @@ class IPPToolWrapper(object):
                          'which_jobs': which_jobs,
                          'my_jobs': my_jobs}}}
         request = get_jobs_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def get_printer_attributes(self, uri,
                                printer_uri=None,
                                requesting_user_name=colander.null,
@@ -469,9 +495,10 @@ class IPPToolWrapper(object):
                          'requesting_user_name': requesting_user_name,
                          'requested_attributes': requested_attributes}}}
         request = get_printer_attributes_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def get_subscriptions(self, uri,
                           printer_uri=None,
                           requesting_user_name=colander.null,
@@ -487,9 +514,10 @@ class IPPToolWrapper(object):
                          'requested_attributes': requested_attributes,
                          'my_subscriptions': my_subscriptions}}}
         request = get_subscriptions_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def get_notifications(self,
                           uri,
                           printer_uri=None,
@@ -505,9 +533,10 @@ class IPPToolWrapper(object):
                    'notify_sequence_numbers': notify_sequence_numbers,
                    'notify_wait': notify_wait}}}
         request = get_notifications_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def cancel_subscription(self, uri,
                             printer_uri=None,
                             requesting_user_name=colander.null,
@@ -518,17 +547,18 @@ class IPPToolWrapper(object):
                 'requesting_user_name': requesting_user_name,
                 'notify_subscription_id': notify_subscription_id}}}
         request = cancel_subscription_form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
+    @pyipptool_coroutine
     def _pause_or_resume_printer(self, form, uri, printer_uri=None,
                                  requesting_user_name=colander.null):
         kw = {'header': {'operation_attributes':
                         {'printer_uri': printer_uri,
                          'requesting_user_name': requesting_user_name}}}
         request = form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
     def pause_printer(self, *args, **kw):
         return self._pause_or_resume_printer(pause_printer_form, *args, **kw)
@@ -536,6 +566,7 @@ class IPPToolWrapper(object):
     def resume_printer(self, *args, **kw):
         return self._pause_or_resume_printer(resume_printer_form, *args, **kw)
 
+    @pyipptool_coroutine
     def _hold_or_release_new_jobs(self, form, uri, printer_uri=None,
                                   requesting_user_name=colander.null,
                                   printer_message_from_operator=colander.null):
@@ -547,8 +578,8 @@ class IPPToolWrapper(object):
                  'printer_message_from_operator': printer_message_from_operator
                  }}}
         request = form.render(kw)
-        response = self._call_ipptool(uri, request)
-        return response['Tests'][0]
+        response = yield self._call_ipptool(uri, request)
+        raise Return(response)
 
     def hold_new_jobs(self, *args, **kw):
         return self._hold_or_release_new_jobs(hold_new_jobs_form, *args, **kw)
@@ -556,3 +587,37 @@ class IPPToolWrapper(object):
     def release_held_new_jobs(self, *args, **kw):
         return self._hold_or_release_new_jobs(release_held_new_jobs_form,
                                               *args, **kw)
+
+
+class AsyncIPPToolWrapper(IPPToolWrapper):
+    async = True
+
+    def __init__(self, config, io_loop):
+        self.config = config
+        self.io_loop = io_loop
+
+    @coroutine
+    def _call_ipptool(self, uri, request):
+        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
+            temp_file.write(request)
+        from tornado.process import Subprocess
+        process = Subprocess([self.config['ipptool_path'],
+                              self.authenticate_uri(uri), '-X',
+                              temp_file.name],
+                             stdin=subprocess.PIPE,
+                             stdout=Subprocess.STREAM,
+                             stderr=Subprocess.STREAM,
+                             io_loop=self.io_loop)
+        future = []
+        self.io_loop.add_timeout(self.io_loop.time() + self.config['timeout'],
+                                 functools.partial(self.timeout_handler,
+                                                   process.proc, future))
+        try:
+            stdout, stderr = yield [Task(process.stdout.read_until_close),
+                                    Task(process.stderr.read_until_close)]
+            if future:
+                raise TimeoutError
+        finally:
+            os.unlink(temp_file.name)
+
+        raise Return(plistlib.readPlistFromString(stdout)['Tests'][0])
